use std::collections::HashMap;

use super::Solution;

use fancy_regex::Regex;

pub struct Day02 {}

impl Solution for Day02 {
    fn part1(
        &self,
        input: &mut Vec<String>,
    ) -> Result<Box<dyn std::fmt::Display + Sync>, Box<dyn std::error::Error>> {
        // Read games
        // Parse into regex with colour to get the number of colours for each game
        // Create an is Game Valid function

        let colour_max: HashMap<&str, u32> =
            HashMap::from([("red", 12), ("green", 13), ("blue", 14)]);

        let mut ans = 0;

        for game in input.iter().enumerate() {
            if self.is_game_valid(game.1, &colour_max)? {
                ans += game.0 + 1
            }
        }
        Ok(Box::new(ans))
    }

    fn part2(
        &self,
        input: &mut Vec<String>,
    ) -> Result<Box<dyn std::fmt::Display + Sync>, Box<dyn std::error::Error>> {
        let mut power = 0u32;
        for game in input.iter().enumerate() {
            let mut min_count: HashMap<&str, u32> =
                HashMap::from([("red", 0), ("green", 0), ("blue", 0)]);

            let hands = game.1.split(';');

            // Look at each time a handful is pulled out
            for run in hands {
                // Find the max number of instances of a given colour in a run
                for colour in &mut min_count {
                    let num_cubes = self.get_num_cubes(run, colour.0)?;
                    if !num_cubes.is_empty() && num_cubes[0] > *colour.1 {
                        //Update with the new max
                        *colour.1 = num_cubes[0];
                    }
                }
            }

            // Calculate the power for this given game
            let mut game_power = 1u32;
            for colour in &min_count {
                game_power *= *colour.1;
            }

            // Combine to formt he total power
            power += game_power;
        }
        Ok(Box::new(power))
    }

    fn get_day(&self) -> u8 {
        2
    }
}

impl Day02 {
    fn is_game_valid(
        &self,
        input: &str,
        colour_max: &HashMap<&'static str, u32>,
    ) -> Result<bool, Box<dyn std::error::Error>> {
        for colour in colour_max {
            let nums = self.get_num_cubes(input, colour.0)?;
            for num in nums {
                if num > *colour.1 {
                    return Ok(false);
                }
            }
        }

        Ok(true)
    }

    fn get_num_cubes(
        &self,
        input: &str,
        colour: &str,
    ) -> Result<Vec<u32>, Box<dyn std::error::Error>> {
        let re = Regex::new(format!("[0-9]+(?= {})", colour).as_str())?;

        // Get the numbers correlating to a given colour and combine
        let nums: Vec<u32> = re
            .find_iter(&input)
            .map(|c| c.unwrap().as_str().parse::<u32>().unwrap_or(0))
            .collect();
        Ok(nums)
    }
}

/// Test from puzzle input
#[cfg(test)]
mod test {
    use super::*;
    use crate::*;

    #[test]
    fn part1() {
        let challenge = day02::Day02 {};

        //Complete the Challenge
        let answer = challenge
            .part1(
                utils::get_input(challenge.get_day(), utils::InputType::Test1)
                    .unwrap()
                    .as_mut(),
            )
            .unwrap()
            .to_string();

        assert_eq!(answer, "8");
    }

    #[test]
    fn part2() {
        let challenge = day02::Day02 {};

        //Complete the Challenge
        let answer = challenge
            .part2(
                utils::get_input(challenge.get_day(), utils::InputType::Test2)
                    .unwrap()
                    .as_mut(),
            )
            .unwrap()
            .to_string();

        assert_eq!(answer, "2286");
    }
}
