use std::collections::HashMap;

use super::Solution;

pub struct Day04 {}

impl Solution for Day04 {
    fn part1(
        &self,
        input: &mut Vec<String>,
    ) -> Result<Box<dyn std::fmt::Display + Sync>, Box<dyn std::error::Error>> {
        // Remove Card XXX: at start
        let cards: Vec<&str> = input
            .iter()
            .map(|c| c.split(":").into_iter().last().unwrap())
            .collect();
        let mut ans = 0;

        for card in cards {
            let winning_nums = self.get_count_winning_numbers(card)?;
            if winning_nums > 0 {
                ans += 2u32.pow((winning_nums - 1) as u32);
            }
        }
        Ok(Box::new(ans))
    }

    fn part2(
        &self,
        input: &mut Vec<String>,
    ) -> Result<Box<dyn std::fmt::Display + Sync>, Box<dyn std::error::Error>> {
        // Remove Card XXX: at start and append card index
        let cards: Vec<(usize, &str)> = input
            .iter()
            .map(|c| c.split(":").into_iter().last().unwrap())
            .into_iter()
            .enumerate()
            .collect();

        let mut copy_counts: Vec<u32> = vec![1; cards.len()];

        for (i, card) in cards.iter() {
            // Create a container for the cards

            // Check if we are looking at a winning card
            let winning_nums = self.get_count_winning_numbers(card)?;
            if winning_nums == 0 {
                continue;
            }

            // We have a winning card... add the next cards resulting * num of current card
            for j in 0..winning_nums {
                // Don't let us add cards that aren't available
                copy_counts[*i + j + 1] += copy_counts[*i];
            }
        }

        Ok(Box::new(copy_counts.iter().sum::<u32>()))
    }

    fn get_day(&self) -> u8 {
        4
    }
}

impl Day04 {
    fn get_count_winning_numbers(&self, card: &str) -> Result<usize, Box<dyn std::error::Error>> {
        let mut winning_numbers: HashMap<u32, bool> = HashMap::new();
        let mut sep_found = false;
        let mut card_nums = vec![];

        for num in card.split_whitespace() {
            // Are we after the |
            if num == "|" {
                sep_found = true;
                continue;
            }
            let num = num.parse::<u32>()?;

            if sep_found {
                if winning_numbers.contains_key(&num) {
                    card_nums.push(num);
                }
            } else {
                winning_numbers.insert(num, true);
            }
        }
        Ok(card_nums.len())
    }
}

/// Test from puzzle input
#[cfg(test)]
mod test {
    use super::*;
    use crate::*;

    #[test]
    fn part1() {
        let challenge = day04::Day04 {};

        //Complete the Challenge
        let answer = challenge
            .part1(
                utils::get_input(challenge.get_day(), utils::InputType::Test1)
                    .unwrap()
                    .as_mut(),
            )
            .unwrap()
            .to_string();

        assert_eq!(answer, "13");
    }

    #[test]
    fn part2() {
        let challenge = day04::Day04 {};

        //Complete the Challenge
        let answer = challenge
            .part2(
                utils::get_input(challenge.get_day(), utils::InputType::Test2)
                    .unwrap()
                    .as_mut(),
            )
            .unwrap()
            .to_string();

        assert_eq!(answer, "30");
    }
}
