use super::Solution;
use itertools::Itertools;

pub struct Day05 {}

impl Solution for Day05 {
    fn part1(
        &self,
        input: &mut Vec<String>,
    ) -> Result<Box<dyn std::fmt::Display + Sync>, Box<dyn std::error::Error>> {
        let (seeds, maps) = self.get_mappings(input)?;
        let locations = maps.iter().fold(seeds, |seeds, mappings| {
            seeds
                .iter()
                .map(|&seed| {
                    mappings
                        .iter()
                        .find(|&&(_, src, range)| (src..src + range).contains(&seed))
                        .map(|(dst, src, _)| dst + seed - src)
                        .unwrap_or(seed)
                })
                .collect()
        });
        Ok(Box::new(*locations.iter().min().unwrap()))
    }

    fn part2(
        &self,
        input: &mut Vec<String>,
    ) -> Result<Box<dyn std::fmt::Display + Sync>, Box<dyn std::error::Error>> {
        let (seeds, maps) = self.get_mappings(input)?;

        let seeds = seeds
            .iter()
            .tuples()
            .map(|(&a, len)| (a, a + len))
            .collect::<Vec<_>>();
        let locations = maps.iter().fold(seeds, |seeds, mappings| {
            seeds
                .iter()
                .flat_map(|&(start, end)| {
                    let mut mapped = Vec::new();
                    let mut unmapped = vec![(start, end)];
                    for &(dst, src, len) in mappings {
                        let mut m = Vec::new();
                        for (start, end) in unmapped {
                            let a = (start, end.min(src));
                            let b = (start.max(src), (src + len).min(end));
                            let c = ((src + len).max(start), end);
                            if a.0 < a.1 {
                                m.push(a);
                            }
                            if b.0 < b.1 {
                                mapped.push((b.0 - src + dst, b.1 - src + dst));
                            }
                            if c.0 < c.1 {
                                m.push(c);
                            }
                        }
                        unmapped = m;
                    }
                    mapped.extend(unmapped);
                    mapped
                })
                .collect()
        });
        Ok(Box::new(locations.iter().map(|&(s, _)| s).min().unwrap()))
    }

    fn get_day(&self) -> u8 {
        5
    }
}

impl Day05 {
    fn get_mappings(
        &self,
        input: &Vec<String>,
    ) -> Result<(Vec<usize>, Vec<Vec<(usize, usize, usize)>>), Box<dyn std::error::Error>> {
        let seeds: Vec<usize> = input[0]
            .split_whitespace()
            .skip(1)
            .map(|s| s.parse().unwrap())
            .collect::<Vec<_>>();

        let maps: Vec<Vec<(usize, usize, usize)>> = input[2..input.len()]
            .join("\n")
            .split("\n\n")
            .map(|s| {
                s.split('\n')
                    .skip(1)
                    .map(|l| {
                        l.split_whitespace()
                            .map(|s| s.parse().unwrap())
                            .collect_tuple()
                            .unwrap()
                    })
                    .collect::<Vec<_>>()
            })
            .collect::<Vec<_>>();

        Ok((seeds, maps))
    }
}

/// Test from puzzle input
#[cfg(test)]
mod test {
    use super::*;
    use crate::*;

    #[test]
    fn part1() {
        let challenge = day05::Day05 {};

        //Complete the Challenge
        let answer = challenge
            .part1(
                utils::get_input(challenge.get_day(), utils::InputType::Test1)
                    .unwrap()
                    .as_mut(),
            )
            .unwrap()
            .to_string();

        assert_eq!(answer, "35");
    }

    #[test]
    fn part2() {
        let challenge = day05::Day05 {};

        //Complete the Challenge
        let answer = challenge
            .part2(
                utils::get_input(challenge.get_day(), utils::InputType::Test2)
                    .unwrap()
                    .as_mut(),
            )
            .unwrap()
            .to_string();

        assert_eq!(answer, "46");
    }
}
