use itertools::Itertools;

use super::Solution;

pub struct Day13 {}

impl Solution for Day13 {
    fn part1(
        &self,
        input: &mut Vec<String>,
    ) -> Result<Box<dyn std::fmt::Display + Sync>, Box<dyn std::error::Error>> {
        let binding = input.clone().iter().join("\n");
        let grids = binding
            .split("\n\n")
            .map(|s| s.split("\n").map(|l| l.as_bytes()).collect::<Vec<_>>())
            .collect::<Vec<_>>();

        Ok(Box::new(self.solve(&grids, 0)))
    }

    fn part2(
        &self,
        input: &mut Vec<String>,
    ) -> Result<Box<dyn std::fmt::Display + Sync>, Box<dyn std::error::Error>> {
        let binding = input.clone().iter().join("\n");
        let grids = binding
            .split("\n\n")
            .map(|s| s.split("\n").map(|l| l.as_bytes()).collect::<Vec<_>>())
            .collect::<Vec<_>>();

        Ok(Box::new(self.solve(&grids, 1)))
    }

    fn get_day(&self) -> u8 {
        13
    }
}

impl Day13 {
    // Finds the index of the column where the reflection is located
    fn find_col_reflection(&self, grid: &[&[u8]], limit: usize) -> Option<usize> {
        (0..grid[0].len() - 1).find(|&c| {
            let incorrect = (0..=c.min(grid[0].len() - c - 2))
                .map(|dc| {
                    let a = c - dc;
                    let b = c + 1 + dc;
                    (0..grid.len())
                        .filter(|&r| grid[r][a] != grid[r][b])
                        .count()
                })
                .sum::<usize>();
            incorrect == limit
        })
    }

    // Finds the index of the row where the reflection is located
    fn find_row_reflection(&self, grid: &[&[u8]], limit: usize) -> Option<usize> {
        (0..grid.len() - 1).find(|&r| {
            let incorrect = (0..=r.min(grid.len() - r - 2))
                .map(|dr| {
                    let a = r - dr;
                    let b = r + 1 + dr;
                    (0..grid[0].len())
                        .filter(|&c| grid[a][c] != grid[b][c])
                        .count()
                })
                .sum::<usize>();
            incorrect == limit
        })
    }

    // Gets the given score for a reflection
    // where limit is the index of the reflection you want to find in each image
    fn solve(&self, grids: &[Vec<&[u8]>], limit: usize) -> usize {
        grids
            .iter()
            .map(|grid| {
                self.find_row_reflection(grid, limit)
                    .map(|r| (r + 1) * 100)
                    .or_else(|| self.find_col_reflection(grid, limit).map(|c| c + 1))
                    .unwrap()
            })
            .sum()
    }
}

/// Test from puzzle input
#[cfg(test)]
mod test {
    use super::*;
    use crate::*;

    #[test]
    fn part1() {
        let challenge = day13::Day13 {};

        //Complete the Challenge
        let answer = challenge
            .part1(
                utils::get_input(challenge.get_day(), utils::InputType::Test1)
                    .unwrap()
                    .as_mut(),
            )
            .unwrap()
            .to_string();

        assert_eq!(answer, "405");
    }

    #[test]
    fn part2() {
        let challenge = day13::Day13 {};

        //Complete the Challenge
        let answer = challenge
            .part2(
                utils::get_input(challenge.get_day(), utils::InputType::Test2)
                    .unwrap()
                    .as_mut(),
            )
            .unwrap()
            .to_string();

        assert_eq!(answer, "400");
    }
}
