use std::{
    error::Error,
    fs::File,
    io::{prelude::*, BufReader},
};

/// Enum used to specify input
#[allow(unused)]
pub enum InputType {
    Test1,
    Test2,
    Actual,
}

/// Function to get the input for a given day of AoC
pub fn get_input(day: u8, input: InputType) -> Result<Vec<String>, Box<dyn Error>> {
    // Find Input File Name
    let file_name = match input {
        InputType::Test1 => format!("src/input/day{:02}_test1", day),
        InputType::Test2 => format!("src/input/day{:02}_test2", day),
        InputType::Actual => format!("src/input/day{:02}", day),
    };

    // Read file into buffer
    let file = File::open(file_name)?;
    let reader = BufReader::new(file);

    // Convert itterate buffer and return string vec
    let data = reader
        .lines()
        .map(|s| s.unwrap_or(String::from("")))
        .collect();

    Ok(data)
}

pub fn lcm(first: u64, second: u64) -> u64 {
    first * second / gcd(first, second)
}

pub fn gcd(first: u64, second: u64) -> u64 {
    let mut max = first;
    let mut min = second;
    if min > max {
        let val = max;
        max = min;
        min = val;
    }

    loop {
        let res = max % min;
        if res == 0 {
            return min;
        }

        max = min;
        min = res;
    }
}
